from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar, Literal, Protocol, Self, cast, final, override, runtime_checkable

import numpy as np
import numpy.typing as npt
import zarr
from numcodecs.pickles import Pickle  # pyright: ignore[reportMissingTypeStubs]
from numpy._typing import _SupportsArray as SupportsArray  # pyright: ignore[reportPrivateUsage]
from zarr.abc.codec import ArrayBytesCodec
from zarr.core.array_spec import ArraySpec
from zarr.core.buffer import Buffer, NDBuffer
from zarr.core.common import JSON, parse_named_configuration
from zarr.core.dtype import VariableLengthBytes, data_type_registry
from zarr.core.dtype.common import DTypeConfig_V2
from zarr.registry import register_codec

import ezarr


@dataclass
class ListWrapper:
    lst: list[Any] | SupportsArray[Any]


@runtime_checkable
class SupportsEZRead(Protocol):
    @classmethod
    def __ez_read__(cls, values: ezarr.EZDict[Any]) -> Self: ...


@runtime_checkable
class SupportsEZWrite(Protocol):
    def __ez_write__(self, values: ezarr.EZDict[Any]) -> None: ...


@runtime_checkable
class SupportsEZReadWrite(SupportsEZRead, SupportsEZWrite, Protocol): ...


@runtime_checkable
class DeferredCreationFunc(Protocol):
    def __call__(self, grp: zarr.Group, name: str, overwrite: bool) -> None: ...


class PyObjectJSON_V2(DTypeConfig_V2[Literal["|O"], Literal["vlen-bytes"]]):
    """
    A wrapper around the JSON representation of the ``PyObject`` data type in Zarr V2.

    The ``name`` field of this class contains the value that would appear under the
    ``dtype`` field in Zarr V2 array metadata. The ``object_codec_id`` field is always ``py-object``

    References
    ----------
    The structure of the ``name`` field is defined in the Zarr V2
    `specification document <https://github.com/zarr-developers/zarr-specs/blob/main/docs/v2/v2.0.rst#data-type-encoding>`__.

    Examples
    --------
    .. code-block:: python

        {
            "name": "|O",
            "object_codec_id": "py-object"
        }
    """


@final
@dataclass(frozen=True, kw_only=True)
class PyObject(VariableLengthBytes):
    """
    A Zarr data type for arrays containing bytes representations of arbitrary Python objects.

    Wraps the NumPy "object" data type. Scalars for this data type are instances of ``bytes``.

    Attributes
    ----------
    dtype_cls: ClassVar[type[np.dtypes.ObjectDType]] = np.dtypes.ObjectDType
        The NumPy data type wrapped by this ZDType.
    _zarr_v3_name: ClassVar[Literal["python_object"]] = "python_object"
        The name of this data type in Zarr V3.
    object_codec_id: ClassVar[Literal["py-object"]] = "py-object"
        The object codec ID for this data type.

    Notes
    -----
    Because this data type uses the NumPy "object" data type, it does not guarantee a compact memory
    representation of array data. Therefore a "vlen-bytes" codec is needed to ensure that the array
    data can be persisted to storage.
    """

    dtype_cls = np.dtypes.ObjectDType
    _zarr_v3_name: ClassVar[Literal["python_object"]] = "python_object"  # pyright: ignore[reportIncompatibleVariableOverride]
    object_codec_id: ClassVar[Literal["py-object"]] = "py-object"  # pyright: ignore[reportIncompatibleVariableOverride]


_pickle = Pickle()


@dataclass(frozen=True)
class PyObjectCodec(ArrayBytesCodec):
    @override
    @classmethod
    def from_dict(cls, data: dict[str, JSON]) -> Self:
        _, configuration_parsed = parse_named_configuration(data, "py-object", require_configuration=False)
        configuration_parsed = configuration_parsed or {}
        return cls(**configuration_parsed)

    @override
    def to_dict(self) -> dict[str, JSON]:
        return {"name": "py-object", "configuration": {}}

    @override
    def evolve_from_array_spec(self, array_spec: ArraySpec) -> Self:
        return self

    @override
    async def _decode_single(self, chunk_data: Buffer, chunk_spec: ArraySpec) -> NDBuffer:
        assert isinstance(chunk_data, Buffer)

        raw_bytes = chunk_data.as_array_like()
        decoded = cast(npt.NDArray[np.object_], _pickle.decode(raw_bytes))
        assert decoded.dtype == np.object_

        decoded.shape = chunk_spec.shape
        return chunk_spec.prototype.nd_buffer.from_numpy_array(decoded)

    @override
    async def _encode_single(self, chunk_data: NDBuffer, chunk_spec: ArraySpec) -> Buffer | None:
        assert isinstance(chunk_data, NDBuffer)
        return chunk_spec.prototype.buffer.from_bytes(_pickle.encode(self.flatten(chunk_data.as_ndarray_like())))

    flatten = np.vectorize(  # pyright: ignore[reportUnannotatedClassAttribute]
        lambda v: v[()].lst  # pyright: ignore[reportAttributeAccessIssue, reportUnknownArgumentType, reportUnknownLambdaType]
        if isinstance(v, np.ndarray) and isinstance(v[()], ListWrapper)
        else v.lst
        if isinstance(v, ListWrapper)
        else v,
        otypes=[np.object_],
    )

    @override
    def compute_encoded_size(self, input_byte_length: int, chunk_spec: ArraySpec) -> int:
        # what is input_byte_length for an object dtype?
        raise NotImplementedError("compute_encoded_size is not implemented for VLen codecs")


register_codec("python_object", PyObjectCodec)
data_type_registry.register("python_object", PyObject)



from typing import Any

import numpy as np
from pytest import fixture

import ezarr


@fixture
def ez_dict():
    data = {"a": 1, "b": [1, 2, 3], "c": {"d": "test", "e": np.arange(100)}, "f": np.zeros((10, 10, 10))}

    return ezarr.EZDict[Any].from_dict(data, store={}, name="data")



from typing import Any, Callable, override

import numpy as np
from pandas._typing import ArrayLike, FloatFormatType

class _GenericArrayFormatter:
    def __init__(
        self,
        values: ArrayLike,
        digits: int = 7,
        formatter: Callable[..., Any] | None = None,
        na_rep: str = "NaN",
        space: str | int = 12,
        float_format: FloatFormatType | None = None,
        justify: str = "right",
        decimal: str = ".",
        quoting: int | None = None,
        fixed_width: bool = True,
        leading_space: bool | None = True,
        fallback_formatter: Callable[..., Any] | None = None,
    ) -> None: ...
    def get_result(self) -> list[str]: ...
    def _format_strings(self) -> list[str]: ...

class FloatArrayFormatter(_GenericArrayFormatter):
    def __init__(
        self,
        values: ArrayLike,
        digits: int = 7,
        formatter: Callable[..., Any] | None = None,
        na_rep: str = "NaN",
        space: str | int = 12,
        float_format: FloatFormatType | None = None,
        justify: str = "right",
        decimal: str = ".",
        quoting: int | None = None,
        fixed_width: bool = True,
        leading_space: bool | None = True,
        fallback_formatter: Callable[..., Any] | None = None,
    ) -> None: ...
    def _value_formatter(
        self, float_format: FloatFormatType | None = None, threshold: float | None = None
    ) -> Callable[..., Any]: ...
    def get_result_as_array(self) -> np.ndarray: ...
    @override
    def _format_strings(self) -> list[str]: ...



from typing import Any, Self, override
from pandas.core.accessor import DirNamesMixin

class PandasObject(DirNamesMixin):
    """
    Baseclass for various pandas objects.
    """

    _cache: dict[str, Any]

    @property
    def _constructor(self) -> Self: ...
    @override
    def __repr__(self) -> str: ...
    def _reset_cache(self, key: str | None = None) -> None: ...
    @override
    def __sizeof__(self) -> int: ...



from typing import final, override

class DirNamesMixin:
    _accessors: set[str] = set()
    _hidden_attrs: frozenset[str] = frozenset()

    @final
    def _dir_deletions(self) -> set[str]: ...
    def _dir_additions(self) -> set[str]: ...
    @override
    def __dir__(self) -> list[str]: ...



from typing import final


@final
class Attribute:
    EZType = "__ez_type__"
    EZClass = "__ez_class__"

    EZVectorSize = "size"


@final
class EZType:
    Object = "object"
    List = "list"
    Vector = "vector"


UNKNOWN = "<UNKNWON>"



import zarr
import numpy as np

import ezarr.inplace as ezi


def test_delete():
    array = zarr.create_array({}, data=np.array([[1, 2, 3], [4, 5, 6]]))
    assert np.array_equal(np.array(ezi.delete(array, 1, axis=1)), np.array([[1, 3], [4, 6]]))


def test_insert():
    array = zarr.create_array({}, data=np.array([[1, 2, 3], [4, 5, 6]]))
    assert np.array_equal(np.array(ezi.insert(array, 1, [7, 8], axis=1)), np.array([[1, 7, 2, 3], [4, 8, 5, 6]]))


def test_add():
    array = zarr.create_array({}, data=np.array([[1, 2, 3], [4, 5, 6]]))
    ezi.add(array, 1)

    assert np.array_equal(np.array(array), np.array([[2, 3, 4], [5, 6, 7]]))



from __future__ import annotations

from contextlib import AbstractContextManager, contextmanager
import itertools as it
from pathlib import Path
import warnings
from collections.abc import ItemsView, Iterator, Mapping, MutableMapping
from typing import TYPE_CHECKING, Any, Literal, NotRequired, Self, TypedDict, cast, override

import numpy as np
import numpy.typing as npt
import zarr
from numpy._typing import _SupportsArray as SupportsArray  # pyright: ignore[reportPrivateUsage]
from zarr.core.array import DEFAULT_FILL_VALUE, CompressorsLike, FiltersLike, SerializerLike, ShardsLike
from zarr.core.array_spec import ArrayConfigLike
from zarr.core.attributes import Attributes
from zarr.core.chunk_key_encodings import ChunkKeyEncodingLike
from zarr.core.common import JSON, AccessModeLiteral, DimensionNames, MemoryOrder, ShapeLike
from zarr.core.dtype import ZDTypeLike
from zarr.errors import UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

import ezarr
from ezarr import io
from ezarr._repr import repr_element
from ezarr.object import EZObject

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem

type DictData[T] = Mapping[str, int | float | np.integer | np.floating | list[Any] | npt.ArrayLike | DictData[T]]
type GroupItems = ItemsView[str, zarr.Group | zarr.Array]


class DictParameters(TypedDict):
    compressors: NotRequired[CompressorsLike]


@contextmanager
def ParametersManager(ez_dict: EZDict[Any], parameters: DictParameters):
    original_parameters = ez_dict.get_parameters()
    ez_dict.set_parameters(parameters)
    try:
        yield

    finally:
        ez_dict.set_parameters(original_parameters)


class EZDict[T](EZObject[dict[str, T]], MutableMapping[str, T]):
    """
    Dict-like object wrapping a zarr.Group for storing arbitrary Python objects

    Args:
        group: a zarr.Group

    Example:
        >>> EZDict(zarr.open_group({}))
        EZDict{}
    """

    __match_args__: tuple[str] = ("_group",)

    def __init__(self, group: zarr.Group) -> None:
        self._group: zarr.Group = group

        self._parameters: DictParameters = DictParameters()

    @classmethod
    def from_dict(
        cls,
        dct: Mapping[str, Any],
        *,
        name: str = "",
        store: StoreLike | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
        attributes: dict[str, JSON] | None = None,
        parameters: DictParameters | None = None,
    ) -> Self:
        r"""
        Create an EZDict from a regular in-memory Python dictionary.

        Args:
            dct: dictionary with arbitrary data to store.
            name: name for the EZDict, to use inside the store.
            store: Store or path to directory in file system or nam of zip file.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: Group path within store.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)
            attributes: (optional) a dictionary of user-defined attributes.
            parameters: (optional) parameters passed during array creation.

        Example:
            >>> data = {"a": 1, "b": [1, 2, 3], "c": {"d": "some text"}}
            >>> ez_dict = EZDict.from_dict(data)
            >>> repr(ez_dict)
            'EZDict{\n\ta: 1,\n\tb: [1 2 3],\n\tc: {...}\n}'
        """
        if store is None:
            store = {}

        grp = zarr.open_group(store, mode=mode, path=path)

        if name:
            grp = grp.create_group(name, overwrite=overwrite)

        if attributes is not None:
            grp.attrs.update(attributes)

        to_visit: list[tuple[Any, str, str]] = [(dct[k], k, "") for k in dct.keys()]

        while len(to_visit):
            value, name_, path_ = to_visit.pop()

            if isinstance(value, Mapping):
                to_visit.extend([(value[k], k, f"{path_}/{name_}" if path_ else name_) for k in value.keys()])  # pyright: ignore[reportUnknownArgumentType, reportUnknownVariableType]

            elif isinstance(value, list | SupportsArray):
                io.write_object(
                    grp,
                    obj=np.asarray(value),  # pyright: ignore[reportUnknownArgumentType]
                    name=name_,
                    path=path_,
                    overwrite=overwrite,
                    parameters=parameters,
                )

            else:
                io.write_object(grp, obj=value, name=name_, path=path_, overwrite=overwrite, parameters=parameters)

        return cls(grp)

    @staticmethod
    def _repr(grp: zarr.Group) -> str:
        if not len(grp):
            return "{}"

        if len(grp) > 100:
            return (
                "{\n\t"
                + ",\n\t".join(
                    [
                        f"{name}: {repr_element(io.read_object(grp, name=name), prefix=f'\t{" " * len(name)}  ')}"
                        for name in it.islice(sorted(grp.keys()), 0, 10)
                    ]
                )
                + ",\n\t...,\n\t"
                + ",\n\t".join(
                    [
                        f"{name}: {repr_element(io.read_object(grp, name=name), prefix=f'\t{" " * len(name)}  ')}"
                        for name in it.islice(sorted(grp.keys()), len(grp) - 10, None)
                    ]
                )
                + "}"
            )

        return (
            "{\n\t"
            + ",\n\t".join(
                [
                    f"{name}: {repr_element(io.read_object(grp, name=name), prefix=f'\t{" " * len(name)}  ')}"
                    for name in sorted(grp.keys())
                ]
            )
            + "\n}"
        )

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}{self._repr(self._group)}"

    @override
    def __len__(self) -> int:
        return len(self._group)

    @override
    def __getitem__(self, key: str, /) -> T:
        return io.read_object(self._group, name=key)

    def __matmul__(self, key: str) -> EZDict[T] | zarr.Array:
        """Get raw Group/Array objects from this EZDict"""
        res = self._group[key]

        if isinstance(res, zarr.Group):
            return EZDict(res)

        return res

    @override
    def __setitem__(self, key: str, value: T, /) -> None:
        if key in self:
            try:
                # try to compare values, might fail when comparing arrays with different shapes
                if self[key] == value:
                    return

            except ValueError:
                pass

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            io.write_object(self._group, obj=value, name=key, overwrite=True, parameters=self._parameters)

    @override
    def __delitem__(self, key: str, /) -> None:
        del self._group[key]

    @override
    def __iter__(self) -> Iterator[str]:
        yield from sorted(self._group.keys())

    def __deepcopy__(self, _memo: dict[Any, Any]) -> dict[str, Any]:
        return self.copy()

    def __ior__(self, other: object) -> EZDict[T]:
        if not isinstance(other, Mapping):
            raise NotImplementedError

        other = cast(Mapping[str, T], other)
        for name, value in other.items():
            if isinstance(value, Mapping):
                grp = self._group.require_group(name)
                EZDict(grp).__ior__(value)  # pyright: ignore[reportUnknownArgumentType]

            else:
                with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                    self[name] = value

        return self

    @property
    def attrs(self) -> Attributes:
        return self._group.attrs

    @property
    def group(self) -> zarr.Group:
        return self._group

    def get_parameters(self) -> DictParameters:
        return self._parameters

    def set_parameters(self, params: DictParameters) -> None:
        self._parameters = params

    def parameters(self, compressors: CompressorsLike = None) -> AbstractContextManager[None]:
        return ParametersManager(self, DictParameters(compressors=compressors))

    def create_array(
        self,
        name: str,
        *,
        shape: ShapeLike | None = None,
        dtype: ZDTypeLike | None = None,
        data: np.ndarray[Any, np.dtype[Any]] | None = None,
        chunks: tuple[int, ...] | Literal["auto"] = "auto",
        shards: ShardsLike | None = None,
        filters: FiltersLike = "auto",
        compressors: CompressorsLike = "auto",
        serializer: SerializerLike = "auto",
        fill_value: Any | None = DEFAULT_FILL_VALUE,
        order: MemoryOrder | None = None,
        attributes: dict[str, JSON] | None = None,
        chunk_key_encoding: ChunkKeyEncodingLike | None = None,
        dimension_names: DimensionNames = None,
        storage_options: dict[str, Any] | None = None,
        overwrite: bool = False,
        config: ArrayConfigLike | None = None,
        write_data: bool = True,
    ) -> zarr.Array:
        """Direct array creation interface.

        Args:
            name : str
                The name of the array relative to the group. If ``path`` is ``None``, the array will be located
                at the root of the store.
            shape : ShapeLike, optional
                Shape of the array. Must be ``None`` if ``data`` is provided.
            dtype : npt.DTypeLike | None
                Data type of the array. Must be ``None`` if ``data`` is provided.
            data : Array-like data to use for initializing the array. If this parameter is provided, the
                ``shape`` and ``dtype`` parameters must be ``None``.
            chunks : tuple[int, ...], optional
                Chunk shape of the array.
                If not specified, default are guessed based on the shape and dtype.
            shards : tuple[int, ...], optional
                Shard shape of the array. The default value of ``None`` results in no sharding at all.
            filters : Iterable[Codec] | Literal["auto"], optional
                Iterable of filters to apply to each chunk of the array, in order, before serializing that
                chunk to bytes.

                For Zarr format 3, a "filter" is a codec that takes an array and returns an array,
                and these values must be instances of :class:`zarr.abc.codec.ArrayArrayCodec`, or a
                dict representations of :class:`zarr.abc.codec.ArrayArrayCodec`.

                For Zarr format 2, a "filter" can be any numcodecs codec; you should ensure that the
                the order if your filters is consistent with the behavior of each filter.

                The default value of ``"auto"`` instructs Zarr to use a default used based on the data
                type of the array and the Zarr format specified. For all data types in Zarr V3, and most
                data types in Zarr V2, the default filters are empty. The only cases where default filters
                are not empty is when the Zarr format is 2, and the data type is a variable-length data type like
                :class:`zarr.dtype.VariableLengthUTF8` or :class:`zarr.dtype.VariableLengthUTF8`. In these cases,
                the default filters contains a single element which is a codec specific to that particular data type.

                To create an array with no filters, provide an empty iterable or the value ``None``.
            compressors : Iterable[Codec], optional
                List of compressors to apply to the array. Compressors are applied in order, and after any
                filters are applied (if any are specified) and the data is serialized into bytes.

                For Zarr format 3, a "compressor" is a codec that takes a bytestream, and
                returns another bytestream. Multiple compressors my be provided for Zarr format 3.
                If no ``compressors`` are provided, a default set of compressors will be used.
                These defaults can be changed by modifying the value of ``array.v3_default_compressors``
                in :mod:`zarr.core.config`.
                Use ``None`` to omit default compressors.

                For Zarr format 2, a "compressor" can be any numcodecs codec. Only a single compressor may
                be provided for Zarr format 2.
                If no ``compressor`` is provided, a default compressor will be used.
                in :mod:`zarr.core.config`.
                Use ``None`` to omit the default compressor.
            serializer : dict[str, JSON] | ArrayBytesCodec, optional
                Array-to-bytes codec to use for encoding the array data.
                Zarr format 3 only. Zarr format 2 arrays use implicit array-to-bytes conversion.
                If no ``serializer`` is provided, a default serializer will be used.
                These defaults can be changed by modifying the value of ``array.v3_default_serializer``
                in :mod:`zarr.core.config`.
            fill_value : Any, optional
                Fill value for the array.
            order : {"C", "F"}, optional
                The memory of the array (default is "C").
                For Zarr format 2, this parameter sets the memory order of the array.
                For Zarr format 3, this parameter is deprecated, because memory order
                is a runtime parameter for Zarr format 3 arrays. The recommended way to specify the memory
                order for Zarr format 3 arrays is via the ``config`` parameter, e.g. ``{'config': 'C'}``.
                If no ``order`` is provided, a default order will be used.
                This default can be changed by modifying the value of ``array.order`` in :mod:`zarr.core.config`.
            attributes : dict, optional
                Attributes for the array.
            chunk_key_encoding : ChunkKeyEncoding, optional
                A specification of how the chunk keys are represented in storage.
                For Zarr format 3, the default is ``{"name": "default", "separator": "/"}}``.
                For Zarr format 2, the default is ``{"name": "v2", "separator": "."}}``.
            dimension_names : Iterable[str], optional
                The names of the dimensions (default is None).
                Zarr format 3 only. Zarr format 2 arrays should not use this parameter.
            storage_options : dict, optional
                If using an fsspec URL to create the store, these will be passed to the backend implementation.
                Ignored otherwise.
            overwrite : bool, default False
                Whether to overwrite an array with the same name in the store, if one exists.
            config : ArrayConfig or ArrayConfigLike, optional
                Runtime configuration for the array.
            write_data : bool
                If a pre-existing array-like object was provided to this function via the ``data`` parameter
                then ``write_data`` determines whether the values in that array-like object should be
                written to the Zarr array created by this function. If ``write_data`` is ``False``, then the
                array will be left empty.
        """
        return self._group.create_array(
            name,
            shape=shape,
            dtype=dtype,
            data=data,
            chunks=chunks,
            shards=shards,
            filters=filters,
            compressors=compressors,
            serializer=serializer,
            fill_value=fill_value,
            order=order,
            attributes=attributes,
            chunk_key_encoding=chunk_key_encoding,
            dimension_names=dimension_names,
            storage_options=storage_options,
            overwrite=overwrite,
            config=config,
            write_data=write_data,
        )

    def put(self, values: SupportsKeysAndGetItem[str, T]) -> None:
        self.clear()
        self.update(values)

    @classmethod
    @override
    def open(
        cls, store: StoreLike | None = None, *, name: str = "", mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        r"""
        Open this EZDict from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZDict, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.

        Example:
            >>> ez_dict = EZDict.from_dict({"a": 1, "b": [1, 2, 3], "c": {"d": "some text"}})
            >>> ez_dict.save("/tmp/dict", overwrite=True)
            >>> repr(ez_dict.open("/tmp/dict"))
            'EZDict{\n\ta: 1,\n\tb: [1 2 3]\n}'
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if isinstance(store, str | Path):
            store = Path(store).expanduser()

        return cls(zarr.open_group(store, mode=mode, path=path))

    @override
    def save(
        self,
        store: StoreLike,
        *,
        name: str = "",
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this EZDict to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZDict, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZDict will be saved.
            overwrite: overwrite EZDict if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> ez_dict = EZDict.from_dict({"a": 1, "b": [1, 2, 3], "c": {"d": "some text"}})
            >>> ez_dict.save("/tmp/dict", overwrite=True)
            >>> Path("/tmp/dict").exists()
            True
        """
        path = f"{path.rstrip('/')}/{name}/" if path else f"{name}/"

        for _, array in self._group.arrays():
            assert isinstance(array, zarr.Array)
            zarr.create_array(store, name=path + array.path, data=array, overwrite=overwrite)  # pyright: ignore[reportArgumentType]

    @staticmethod
    def _copy_nested(dct: dict[str, Any], value: EZDict[Any]) -> dict[str, Any]:
        for k, v in value.items():
            if isinstance(v, EZDict):
                sub = dct.setdefault(k, {})
                EZDict._copy_nested(sub, v)  # pyright: ignore[reportUnknownArgumentType]

            elif isinstance(v, ezarr.EZList):
                dct[k] = v.copy()

            elif isinstance(v, zarr.Array):
                dct[k] = np.array(v)

            else:
                dct[k] = v

        return dct

    @override
    def copy(self) -> dict[str, T]:
        """
        Convert this EZDict into a Python dict, loading all the data into memory.

        Example:
            >>> ez_dict = EZDict.from_dict({"a": 1, "b": [1, 2, 3], "c": {"d": "some text"}})
            >>> ez_dict.copy()
            {'a': np.int64(1), 'b': array([1, 2, 3]), 'c': {'d': np.str_('some text')}}
        """
        return EZDict._copy_nested({}, self)



from typing import Self, final

from pandas._typing import AxisInt, Shape
from pandas.core.base import PandasObject
from pandas.core.indexes.api import Index

class DataManager(PandasObject):
    axes: list[Index]

    @property
    def items(self) -> Index: ...
    @final
    def __len__(self) -> int: ...
    @property
    def ndim(self) -> int: ...
    @property
    def shape(self) -> Shape: ...
    @final
    def _validate_set_axis(self, axis: AxisInt, new_labels: Index) -> None: ...
    def reindex_indexer(
        self,
        new_axis,
        indexer,
        axis: AxisInt,
        fill_value=None,
        allow_dups: bool = False,
        copy: bool = True,
        only_slice: bool = False,
    ) -> Self:
        raise AbstractMethodError(self)

    @final
    def reindex_axis(self, new_index: Index, axis: AxisInt, fill_value=None, only_slice: bool = False) -> Self:
        """
        Conform data manager to new index.
        """
        new_index, indexer = self.axes[axis].reindex(new_index)

        return self.reindex_indexer(
            new_index, indexer, axis=axis, fill_value=fill_value, copy=False, only_slice=only_slice
        )

    def _equal_values(self, other: Self) -> bool:
        """
        To be implemented by the subclasses. Only check the column values
        assuming shape and indexes have already been checked.
        """
        raise AbstractMethodError(self)

    @final
    def equals(self, other: object) -> bool:
        """
        Implementation for DataFrame.equals
        """
        if not isinstance(other, type(self)):
            return False

        self_axes, other_axes = self.axes, other.axes
        if len(self_axes) != len(other_axes):
            return False
        if not all(ax1.equals(ax2) for ax1, ax2 in zip(self_axes, other_axes)):
            return False

        return self._equal_values(other)

    def apply(self, f, align_keys: list[str] | None = None, **kwargs) -> Self:
        raise AbstractMethodError(self)

    def apply_with_block(self, f, align_keys: list[str] | None = None, **kwargs) -> Self:
        raise AbstractMethodError(self)

    @final
    def isna(self, func) -> Self:
        return self.apply("apply", func=func)

    @final
    def fillna(self, value, limit: int | None, inplace: bool, downcast) -> Self:
        if limit is not None:
            # Do this validation even if we go through one of the no-op paths
            limit = libalgos.validate_limit(None, limit=limit)

        return self.apply_with_block(
            "fillna",
            value=value,
            limit=limit,
            inplace=inplace,
            downcast=downcast,
            using_cow=using_copy_on_write(),
            already_warned=_AlreadyWarned(),
        )

    @final
    def where(self, other, cond, align: bool) -> Self:
        if align:
            align_keys = ["other", "cond"]
        else:
            align_keys = ["cond"]
            other = extract_array(other, extract_numpy=True)

        return self.apply_with_block(
            "where", align_keys=align_keys, other=other, cond=cond, using_cow=using_copy_on_write()
        )

    @final
    def putmask(self, mask, new, align: bool = True, warn: bool = True) -> Self:
        if align:
            align_keys = ["new", "mask"]
        else:
            align_keys = ["mask"]
            new = extract_array(new, extract_numpy=True)

        already_warned = None
        if warn_copy_on_write():
            already_warned = _AlreadyWarned()
            if not warn:
                already_warned.warned_already = True

        return self.apply_with_block(
            "putmask",
            align_keys=align_keys,
            mask=mask,
            new=new,
            using_cow=using_copy_on_write(),
            already_warned=already_warned,
        )

    @final
    def round(self, decimals: int, using_cow: bool = False) -> Self:
        return self.apply_with_block("round", decimals=decimals, using_cow=using_cow)

    @final
    def replace(self, to_replace, value, inplace: bool) -> Self:
        inplace = validate_bool_kwarg(inplace, "inplace")
        # NDFrame.replace ensures the not-is_list_likes here
        assert not lib.is_list_like(to_replace)
        assert not lib.is_list_like(value)
        return self.apply_with_block(
            "replace",
            to_replace=to_replace,
            value=value,
            inplace=inplace,
            using_cow=using_copy_on_write(),
            already_warned=_AlreadyWarned(),
        )

    @final
    def replace_regex(self, **kwargs) -> Self:
        return self.apply_with_block(
            "_replace_regex", **kwargs, using_cow=using_copy_on_write(), already_warned=_AlreadyWarned()
        )

    @final
    def replace_list(
        self, src_list: list[Any], dest_list: list[Any], inplace: bool = False, regex: bool = False
    ) -> Self:
        """do a list replace"""
        inplace = validate_bool_kwarg(inplace, "inplace")

        bm = self.apply_with_block(
            "replace_list",
            src_list=src_list,
            dest_list=dest_list,
            inplace=inplace,
            regex=regex,
            using_cow=using_copy_on_write(),
            already_warned=_AlreadyWarned(),
        )
        bm._consolidate_inplace()
        return bm

    def interpolate(self, inplace: bool, **kwargs) -> Self:
        return self.apply_with_block(
            "interpolate", inplace=inplace, **kwargs, using_cow=using_copy_on_write(), already_warned=_AlreadyWarned()
        )

    def pad_or_backfill(self, inplace: bool, **kwargs) -> Self:
        return self.apply_with_block(
            "pad_or_backfill",
            inplace=inplace,
            **kwargs,
            using_cow=using_copy_on_write(),
            already_warned=_AlreadyWarned(),
        )

    def shift(self, periods: int, fill_value) -> Self:
        if fill_value is lib.no_default:
            fill_value = None

        return self.apply_with_block("shift", periods=periods, fill_value=fill_value)

    # --------------------------------------------------------------------
    # Consolidation: No-ops for all but BlockManager

    def is_consolidated(self) -> bool:
        return True

    def consolidate(self) -> Self:
        return self

    def _consolidate_inplace(self) -> None:
        return

class SingleDataManager(DataManager):
    @property
    def ndim(self) -> Literal[1]:
        return 1

    @final
    @property
    def array(self) -> ArrayLike:
        """
        Quick access to the backing array of the Block or SingleArrayManager.
        """
        # error: "SingleDataManager" has no attribute "arrays"; maybe "array"
        return self.arrays[0]  # type: ignore[attr-defined]

    def setitem_inplace(self, indexer, value, warn: bool = True) -> None:
        """
        Set values with indexer.

        For Single[Block/Array]Manager, this backs s[indexer] = value

        This is an inplace version of `setitem()`, mutating the manager/values
        in place, not returning a new Manager (and Block), and thus never changing
        the dtype.
        """
        arr = self.array

        # EAs will do this validation in their own __setitem__ methods.
        if isinstance(arr, np.ndarray):
            # Note: checking for ndarray instead of np.dtype means we exclude
            #  dt64/td64, which do their own validation.
            value = np_can_hold_element(arr.dtype, value)

        if isinstance(value, np.ndarray) and value.ndim == 1 and len(value) == 1:
            # NumPy 1.25 deprecation: https://github.com/numpy/numpy/pull/10615
            value = value[0, ...]

        arr[indexer] = value

    def grouped_reduce(self, func):
        arr = self.array
        res = func(arr)
        index = default_index(len(res))

        mgr = type(self).from_array(res, index)
        return mgr

    @classmethod
    def from_array(cls, arr: ArrayLike, index: Index):
        raise AbstractMethodError(cls)



import tempfile
from collections.abc import Generator

import pandas as pd
import pytest

from ezarr.dataframe import EZDataFrame


@pytest.fixture
def ezdf() -> Generator[EZDataFrame, None, None]:
    with tempfile.TemporaryDirectory() as path:
        ezdf = EZDataFrame(
            pd.DataFrame({"col_int": [1, 2, 3], "col_str": ["a", "bc", "def"], "col_float": [1.5, 2.5, 3.5]})
        )

        ezdf.save(path, name="data")

        yield EZDataFrame.open(path, name="data")



import numpy as np
import pandas as pd

from ezarr.dataframe import EZDataFrame


def test_can_replace_column(ezdf: EZDataFrame) -> None:
    ezdf["col_int"] = [-1, -2, -3]
    assert ezdf["col_int"].equals(pd.Series([-1, -2, -3]))


def test_can_set_column_from_series(ezdf: EZDataFrame) -> None:
    ezdf["new_col"] = pd.Series([-1, -2, -3])
    assert np.array_equal(ezdf.columns, ["col_int", "col_str", "col_float", "new_col"])



from __future__ import annotations

from pathlib import Path
import re
import warnings
from collections.abc import Callable, Generator, Hashable, Iterable, Mapping, MutableMapping, Sequence
from typing import Any, Literal, NoReturn, Self, cast, overload, override

import numpy as np
import numpy.typing as npt
import pandas as pd
import pandas._typing as pdt
import zarr
import zarr.abc.store
import zarr.storage
from pandas.core.generic import NDFrame
from pandas.core.internals.array_manager import BaseArrayManager  # pyright: ignore[reportMissingImports, reportUnknownVariableType]
from zarr.core.common import AccessModeLiteral
from zarr.errors import UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

import ezarr
from ezarr.common import write_array  # pyright: ignore[reportUnknownVariableType]
from ezarr.dataframe.manager import EZArrayManager
from ezarr.types import SupportsEZReadWrite


def _get_arrays(
    store: zarr.Group,
    arrays: Iterable[tuple[Hashable, Any]],
    dtypes: Mapping[Hashable, str | np.dtype | type[str | complex | bool | object]],
) -> Generator[zarr.Array, None, None]:
    for name, arr in arrays:
        dtype = dtypes.get(name)

        if isinstance(dtype, pd.CategoricalDtype):
            yield write_array(arr, store.store, name=str(name), dtype=dtype.categories.dtype, casting="unsafe")  # pyright: ignore[reportArgumentType]

        else:
            yield write_array(arr, store.store, name=str(name), dtype=dtype, casting="unsafe")


class EZDataFrame(pd.DataFrame, SupportsEZReadWrite):  # EZObject[pd.DataFrame],
    """
    Proxy for pandas DataFrames for storing data in zarr stores.
    """

    _internal_names: list[str] = ["_data_store"] + pd.DataFrame._internal_names  # pyright: ignore[reportAttributeAccessIssue]
    _internal_names_set: set[str] = {"_data_store"} | pd.DataFrame._internal_names_set  # pyright: ignore[reportAttributeAccessIssue]

    _data_store: zarr.Group

    @property
    def _constructor(self) -> Callable[..., EZDataFrame]:
        def inner(df: Any, copy: bool = False) -> EZDataFrame:
            if isinstance(df, pd.DataFrame):
                return EZDataFrame(df)

            if not isinstance(df, EZDataFrame):
                raise ValueError("EZDataFrame constructor not properly called")

            if copy:
                raise NotImplementedError

            return df

        return inner

    def _constructor_from_mgr(self, mgr: BaseArrayManager, axes: list[pd.Index]) -> pd.DataFrame:
        if not isinstance(mgr, EZArrayManager):
            df: pd.DataFrame = pd.DataFrame._from_mgr(mgr, axes=axes)  # pyright: ignore[reportAttributeAccessIssue]

        else:
            df = EZDataFrame._from_mgr(mgr, axes=axes)  # pyright: ignore[reportAttributeAccessIssue]

        if isinstance(self, pd.DataFrame):
            # This would also work `if self._constructor is DataFrame`, but
            #  this check is slightly faster, benefiting the most-common case.
            return df

        elif type(self).__name__ == "GeoDataFrame":
            # Shim until geopandas can override their _constructor_from_mgr
            #  bc they have different behavior for Managers than for DataFrames
            return self._constructor(mgr)

        # We assume that the subclass __init__ knows how to handle a
        #  pd.DataFrame object.
        return self._constructor(df)

    def _constructor_sliced_from_mgr(self, mgr: BaseArrayManager, axes: list[pd.Index[Any]]) -> pd.Series[Any]:
        ser: pd.Series = pd.Series._from_mgr(mgr, axes)  # pyright: ignore[reportAttributeAccessIssue]
        ser._name = None  # caller is responsible for setting real name

        if isinstance(self, pd.DataFrame):
            return ser

        return self._constructor_sliced(ser)

    def __init__(
        self,
        data: pd.DataFrame
        | MutableMapping[Hashable, npt.NDArray[Any] | Sequence[Any]]
        | ezarr.EZDict[Any]
        | npt.NDArray[Any]
        | zarr.Array
        | tuple[EZArrayManager, zarr.Group]
        | None = None,
        index: zarr.Array | pd.Index | None = None,
        columns: zarr.Array | pd.Index | None = None,
        dtypes: Mapping[Hashable, str | np.dtype | type[str | complex | bool | object]] | None = None,
    ):
        match data:
            case ezarr.EZDict():
                assert index is None, (
                    "Cannot provide the `index` parameter when also providing the `data` parameter as EZDict."
                )
                assert columns is None, (
                    "Cannot provide the `columns` parameter when also providing the `data` parameter as EZDict."
                )

                _index: pd.Index = pd.Index(data["index"].copy())
                _columns: pd.Index = pd.Index(data["arrays"].keys())
                arrays: list[zarr.Array] = [arr for arr in data["arrays"].values()]

                store = data.group
                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case pd.DataFrame():
                _index = data.index if index is None else pd.Index(np.array(index))
                _columns = data.columns if columns is None else pd.Index(np.array(columns))

                store = zarr.create_group({})
                with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                    arrays = list(
                        _get_arrays(store, data.to_dict(orient="series").items(), dtypes or data.dtypes.to_dict())
                    )

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case np.ndarray() | zarr.Array():
                _index = pd.RangeIndex(start=0, stop=data.shape[0]) if index is None else pd.Index(np.array(index))
                _columns = (
                    pd.RangeIndex(start=0, stop=data.shape[1]) if columns is None else pd.Index(np.array(columns))
                )

                store = zarr.create_group({})
                with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                    arrays = list(_get_arrays(store, zip(_columns, np.transpose(np.asarray(data))), dtypes or {}))

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case MutableMapping():
                assert columns is None, (
                    "Cannot provide the `columns` parameter when also providing the `data` parameter as dict."
                )

                _columns = pd.Index(data.keys())

                store = zarr.create_group({})
                arrays = list(_get_arrays(store, data.items(), dtypes or {}))

                _index = pd.RangeIndex(start=0, stop=arrays[0].shape[0]) if index is None else pd.Index(np.array(index))

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case None:
                _index = pd.RangeIndex(0) if index is None else pd.Index(np.array(index))
                _columns = pd.RangeIndex(0) if columns is None else pd.Index(np.array(columns))
                arrays = []

                store = zarr.create_group({})
                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case (EZArrayManager() as mgr, zarr.Group() as store):
                pass

            case _:  # pyright: ignore[reportUnnecessaryComparison]
                raise TypeError(f"Invalid type '{type(data)}' for 'data' argument.")  # pyright: ignore[reportUnreachable]

        object.__setattr__(self, "_data_store", store)
        NDFrame.__init__(self, mgr)  # pyright: ignore[reportCallIssue]

    @override
    def __finalize__(self, other: EZDataFrame, method: str | None = None, **kwargs: Any) -> pd.DataFrame:  # pyright: ignore[reportIncompatibleMethodOverride]
        super().__finalize__(other, method, **kwargs)

        if method == "copy":
            return other

        return self

    @staticmethod
    def _get_mode(store: zarr.abc.store.Store) -> str:
        match store:
            case zarr.storage.FsspecStore():
                return "FSSpec"
            case zarr.storage.GpuMemoryStore():
                return "GPU"
            case zarr.storage.LocalStore():
                return "Local"
            case zarr.storage.LoggingStore(store=inner_store):  # pyright: ignore[reportUnknownVariableType]
                return f"Logging:{EZDataFrame._get_mode(inner_store)}"  # pyright: ignore[reportUnknownArgumentType]
            case zarr.storage.MemoryStore():
                return "RAM"
            case zarr.storage.ObjectStore():
                return "Object"
            case zarr.storage.WrapperStore(store=inner_store):  # pyright: ignore[reportUnknownVariableType]
                return f"Wrapper:{EZDataFrame._get_mode(inner_store)}"  # pyright: ignore[reportUnknownArgumentType]
            case zarr.storage.ZipStore():
                return "Zip"
            case _:
                return type(store).__name__

    @override
    def __repr__(self) -> str:
        if self.empty:
            return f"""\
Empty {type(self).__name__}
Columns: {self.columns.tolist()}
Index: {self.index.tolist()}
[{self._get_mode(self._data_store.store)}]"""

        repr_ = repr(self.iloc[:5].copy())
        re.sub(r"\n\n\[.*\]$", "", repr_)
        return (
            repr_
            + ("\n..." if self.shape[1] > 5 else "")
            + f"\n[{self._get_mode(self._data_store.store)}]\n[{len(self.index)} rows x {len(self.columns)} columns]"
        )

    @override
    def __ez_write__(self, values: ezarr.EZDict[Any]) -> None:
        if values.group.store == self._data_store.store:
            return

        _index = self.index.values
        if _index.dtype == object:
            _index = _index.astype(str)

        values["index"] = _index
        values["arrays"] = {str(k): v.values for k, v in self.to_dict(orient="series").items()}
        values["arrays"].attrs["columns_order"] = self.columns.to_list()

    @classmethod
    @override
    def __ez_read__(cls, values: ezarr.EZDict[Any]) -> Self:
        arrays = values["arrays"]
        columns_order = values["arrays"].attrs["columns_order"]
        columns_order_str = np.array(columns_order).astype(str)

        mgr = EZArrayManager(
            [
                arr
                for (_, arr) in sorted(
                    arrays.group.arrays(), key=lambda name_arr: np.where(columns_order_str == name_arr[0])[0][0]
                )
            ],
            [pd.Index(values["index"][:]), pd.Index(columns_order)],
            group=values.group,
        )

        return cls((mgr, values.group))

    @property
    def data(self) -> ezarr.EZDict[Any]:
        return ezarr.EZDict(self._data_store)

    @classmethod
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open this object from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if isinstance(store, str | Path):
            store = Path(store).expanduser()

        return cls.__ez_read__(ezarr.EZDict(zarr.open_group(store, mode=mode, path=path)))

    def save(
        self,
        store: StoreLike,
        *,
        name: str,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this object to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the object will be saved.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if not overwrite:
            mode = "w-"

        self.__ez_write__(ezarr.EZDict(zarr.open_group(store, mode=mode, path=path)))

    @overload
    def sort_values(
        self,
        by: str | Sequence[str],
        *,
        axis: pdt.Axis = ...,
        ascending: bool | list[bool] | tuple[bool, ...] = ...,
        inplace: Literal[False] = ...,
        kind: pdt.SortKind = ...,
        na_position: pdt.NaPosition = ...,
        ignore_index: bool = ...,
        key: pdt.ValueKeyFunc = ...,
    ) -> pd.DataFrame: ...

    @overload
    def sort_values(
        self,
        by: str | Sequence[str],
        *,
        axis: pdt.Axis = ...,
        ascending: bool | list[bool] | tuple[bool, ...] = ...,
        inplace: Literal[True],
        kind: pdt.SortKind = ...,
        na_position: pdt.NaPosition = ...,
        ignore_index: bool = ...,
        key: pdt.ValueKeyFunc = ...,
    ) -> NoReturn: ...

    @override
    def sort_values(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        by: str | Sequence[str],
        *,
        axis: pdt.Axis = 0,
        ascending: bool | list[bool] | tuple[bool, ...] = True,
        inplace: bool = False,
        kind: pdt.SortKind = "quicksort",
        na_position: pdt.NaPosition = "last",
        ignore_index: bool = False,
        key: pdt.ValueKeyFunc | None = None,
    ) -> pd.DataFrame:
        """
        Sort by the values along either axis.

        Parameters
        ----------
        by : str or list of str
            Name or list of names to sort by.

            - if `axis` is 0 or `'index'` then `by` may contain index
              levels and/or column labels.
            - if `axis` is 1 or `'columns'` then `by` may contain column
              levels and/or index labels.
        axis : "{0 or 'index', 1 or 'columns'}", default 0
             Axis to be sorted.
        ascending : bool or list of bool, default True
             Sort ascending vs. descending. Specify list for multiple sort
             orders.  If this is a list of bools, must match the length of
             the by.
        inplace : bool, default False
             If True, perform operation in-place.
        kind : {'quicksort', 'mergesort', 'heapsort', 'stable'}, default 'quicksort'
             Choice of sorting algorithm. See also :func:`numpy.sort` for more
             information. `mergesort` and `stable` are the only stable algorithms. For
             DataFrames, this option is only applied when sorting on a single
             column or label.
        na_position : {'first', 'last'}, default 'last'
             Puts NaNs at the beginning if `first`; `last` puts NaNs at the
             end.
        ignore_index : bool, default False
             If True, the resulting axis will be labeled 0, 1, …, n - 1.
        key : callable, optional
            Apply the key function to the values
            before sorting. This is similar to the `key` argument in the
            builtin :meth:`sorted` function, with the notable difference that
            this `key` function should be *vectorized*. It should expect a
            ``Series`` and return a Series with the same shape as the input.
            It will be applied to each column in `by` independently.

        Returns
        -------
        DataFrame or None
            DataFrame with sorted values or None if ``inplace=True``.

        See Also
        --------
        DataFrame.sort_index : Sort a DataFrame by the index.
        Series.sort_values : Similar method for a Series.

        Examples
        --------
        >>> df = pd.DataFrame({
        ...     'col1': ['A', 'A', 'B', np.nan, 'D', 'C'],
        ...     'col2': [2, 1, 9, 8, 7, 4],
        ...     'col3': [0, 1, 9, 4, 2, 3],
        ...     'col4': ['a', 'B', 'c', 'D', 'e', 'F']
        ... })
        >>> df
          col1  col2  col3 col4
        0    A     2     0    a
        1    A     1     1    B
        2    B     9     9    c
        3  NaN     8     4    D
        4    D     7     2    e
        5    C     4     3    F

        Sort by col1

        >>> df.sort_values(by=['col1'])
          col1  col2  col3 col4
        0    A     2     0    a
        1    A     1     1    B
        2    B     9     9    c
        5    C     4     3    F
        4    D     7     2    e
        3  NaN     8     4    D

        Sort by multiple columns

        >>> df.sort_values(by=['col1', 'col2'])
          col1  col2  col3 col4
        1    A     1     1    B
        0    A     2     0    a
        2    B     9     9    c
        5    C     4     3    F
        4    D     7     2    e
        3  NaN     8     4    D

        Sort Descending

        >>> df.sort_values(by='col1', ascending=False)
          col1  col2  col3 col4
        4    D     7     2    e
        5    C     4     3    F
        2    B     9     9    c
        0    A     2     0    a
        1    A     1     1    B
        3  NaN     8     4    D

        Putting NAs first

        >>> df.sort_values(by='col1', ascending=False, na_position='first')
          col1  col2  col3 col4
        3  NaN     8     4    D
        4    D     7     2    e
        5    C     4     3    F
        2    B     9     9    c
        0    A     2     0    a
        1    A     1     1    B

        Sorting with a key function

        >>> df.sort_values(by='col4', key=lambda col: col.str.lower())
          col1  col2  col3 col4
        0    A     2     0    a
        1    A     1     1    B
        2    B     9     9    c
        3  NaN     8     4    D
        4    D     7     2    e
        5    C     4     3    F
        """
        if inplace:
            raise NotImplementedError

        df = cast(pd.DataFrame, self.copy())
        df.sort_values(
            by,
            axis=axis,
            ascending=ascending,
            inplace=True,
            kind=kind,
            na_position=na_position,
            ignore_index=ignore_index,
            key=key,
        )
        return df



import importlib

from ezarr import inplace, io
from ezarr.dict import EZDict
from ezarr.list import EZList
from ezarr.types import PyObject, PyObjectCodec, SupportsEZRead, SupportsEZReadWrite, SupportsEZWrite
from ezarr.vector import EZVector
from ezarr.view import ArrayView

importlib.import_module("ezarr.patch")

__all__ = [
    "ArrayView",
    "EZDict",
    "EZList",
    "EZVector",
    "inplace",
    "io",
    "PyObject",
    "PyObjectCodec",
    "SupportsEZRead",
    "SupportsEZReadWrite",
    "SupportsEZWrite",
]
